-- used to keep track of the number of enemy units
enemyUnits = {}

-- an array of the archer trigger names that we will use to control the triggers, as the player hits the triggers we will remove it from the list so that A, we know how many we have left and B, we know which ones are left so we can manipulate them
triggerList = {"trigger_front_left","trigger_back_left","trigger_back_right","trigger_front_right"}

-- bools used to control messaging
archerUnitUnlocked = false
archerUnitSelected = false
shownFinishThemOffDialogue = false
shownPikemenMovingDialogue = false
shownPlayMilitiaWarningDialogue = false
firstCinematicCameraShown = false

-- called at before the deploy phase
function onDeployStart()
	-- register the triggers in the level
	SetTriggerCallbacks("trigger_front_left", onTriggerEnter_trigger_front_left, onTriggerLeave_trigger_front_left)
	SetTriggerCallbacks("trigger_back_left", onTriggerEnter_trigger_back_left, onTriggerLeave_trigger_back_left)
	SetTriggerCallbacks("trigger_back_right", onTriggerEnter_trigger_back_right, onTriggerLeave_trigger_back_right)
	SetTriggerCallbacks("trigger_front_right", onTriggerEnter_trigger_front_right, onTriggerLeave_trigger_front_right)
	SetTriggerCallbacks("trigger_pikemen_moving", onTriggerEnter_trigger_pikemen_moving, onTriggerLeave_trigger_pikemen_moving)
	SetTriggerCallbacks("trigger_militia_attack", onTriggerEnter_trigger_militia_attack, onTriggerLeave_trigger_militia_attack)
end

-- called after the deploy phase
function onLevelStart()
	
	-- get the initial list of units for us to track
	enemyUnits = GetUnits("*", "*", 1)
	
	-- Set the archers as allied so you can't accidentally attack them (side 0 is player, side 2 is archers)
	SetAlliedPlayers(0, 2)
	-- show the start dialog

end

function onCameraAnimationFinish(cameraName)
	if cameraName == "BATTLE_START" then
		ShowDialog("talkingheads/full/guard","We have cornered a force of pikemen.Use the nearby archers to thin their numbers from afar.")
	end
end
 
function onCinematicCameraEnd()
	if not firstCinematicCameraShown then
		-- put up a message explaining what's going on
		AddTimer(1,callApproachArcherMessage)
		firstCinematicCameraShown = true
	end
end

function callApproachArcherMessage()
	-- put up a message explaining what's going on and clear it after x seconds
	ShowDialog("talkingheads/full/guard","Approach the archers to have them join us")
	
	-- turn on an arrow pointing at the first trigger in the trigger list
	ShowGuideArrowForTrigger(triggerList[1])
	AddTargetCircle(triggerList[1])
end

function onUnitSelected(selectedUnit)
	
	-- if the player selects the archer units
	if archerUnitUnlocked and not archerUnitSelected then
	
		--if the ID contains the text "archer"
		if string.match(selectedUnit.spawnId,"[start]") then
			
			-- make sure we never call this stuff again
			archerUnitSelected = true
			
			-- send a message to unlock the remaining units
			AddTimer(4,getOtherArchersMessage)
			
			--HideUnitGuideArrow(selectedUnit)
			
		end
	end
end

function getOtherArchersMessage()

	-- send message about getting more archers
	ShowDialog("talkingheads/full/guard","Send troops to the other archers to have them join us")
	
	-- put in move orders for all remaining archer units
	for i,trigger in pairs(triggerList) do
		AddTargetCircle(trigger)
	end
	
end

function onUnitDestroyed(deadUnit)
	-- if an enemy unit has died then remove it from the enemy unit array
	-- if the enemy forces drop below a set amount then trigger the clean up message
	if deadUnit.side == 1 then
		-- for each unit in the enemyUnits array
		for i,unit in pairs(enemyUnits) do
			-- if this is the dead unit then remove it from the array
			if unit == deadUnit then
				-- remove the element from the array at point i
				table.remove(enemyUnits,i)
				
				-- if the enemy's army has dropped below 3 then kick off this message
				if #enemyUnits <= 3 and not shownFinishThemOffDialogue then
					-- make sure we don't call this stuff again
					shownFinishThemOffDialogue = true
					
					-- show dialog about finishing them off
					ShowDialog("talkingheads/full/guard", "They are weakened! Attack their exposed flanks and press home our advantage")
				end
				
				-- leave the for loop (because we've found the right unit)
				break
			end
		end
	end
end

-- a dialogue message to warn the player if the pikemen start moving to attack the archers
function onTriggerLeave_trigger_pikemen_moving(unit)
	if unit.side == 1 then
		-- if we've unlocked an archer unit but not shown this message
		if archerUnitUnlocked and not shownPikemenMovingDialogue then
		
			-- disable this trigger so this stuff is not called again
			DisableTrigger("trigger_pikemen_moving")
			
			-- also make sure this stuff is not called again (double redundancy :D)
			shownPikemenMovingDialogue = true
		
			-- show dialogue text about protecting archers
			ShowDialog("talkingheads/full/guard","They are moving to attack! \n\n We should move our militia units in front of the archers to protect them from the enemy pikemen")
			
		end
	end
end

-- dialogue to play if the player tries to attack with miltia before using an archer
function onTriggerEnter_trigger_militia_attack(unit)
	-- if it's a player unit
	if unit.side == 0 then
		-- if we're in the right part of the mission (before the player has used an archer unit)
		if not archerUnitSelected and not shownPlayMilitiaWarningDialogue then
			
		-- disable this trigger so this stuff is not called again
			DisableTrigger("trigger_militia_attack")
			
			-- also make sure this stuff is not called again (double redundancy :D)
			shownPlayMilitiaWarningDialogue = true
			
			-- dialogue about not attacking with milita before using archers
			ShowDialog("talkingheads/full/guard","The pikemen will tear the men apart if they get too close. \n\n Command the archer units nearby to attack from afar")
		end
	end
end

-- turn on the archers if the player gets close to them
function onTriggerEnter_trigger_front_left(unit)
	-- call this function (passing in the unit and the position) to activate the archers, messages and turn off arrows etc
	activateArcherUnit(unit,"front_left")
end

-- turn on the archers if the player gets close to them
function onTriggerEnter_trigger_back_left(unit)
	-- call this function (passing in the unit and the position) to activate the archers, messages and turn off arrows etc
	activateArcherUnit(unit,"back_left")
end

-- turn on the archers if the player gets close to them
function onTriggerEnter_trigger_back_right(unit)
	-- call this function (passing in the unit and the position) to activate the archers, messages and turn off arrows etc
	activateArcherUnit(unit,"back_right")
end

-- turn on the archers if the player gets close to them
function onTriggerEnter_trigger_front_right(unit)
	-- call this function (passing in the unit and the position) to activate the archers, messages and turn off arrows etc
	activateArcherUnit(unit,"front_right")
end

-- this is called by the trigger functions (they all use the same bit of script)
function activateArcherUnit(unit,positionIdentifier)

	-- if it's a friendly unit that is not one of the starting archers
	if unit.side == 0 then
		-- disable the trigger this came from so we don't activate this unit again
		DisableTrigger("trigger_" .. positionIdentifier)
		
		-- find the unit we want to change sides
		archerUnit = GetUnits("archer_" .. positionIdentifier .. "_[start]")
		
		-- Switch the unit's allegiance to the player's side 
		SwitchSideToPlayer(archerUnit[1])
		
		-- tell the unit to hold ground
		SetUnitStandGuard(archerUnit[1],true)
		
		-- remove the trigger from the triggerList
		removeTriggerFromTriggerList("trigger_" .. positionIdentifier)
		
		-- clear all of the guide arrows
		-- clear the guide arrows
		--ClearAllGuideArrows()
		
		-- clear all of the move to circles
		ClearTargetCircles()
		
		-- if we've not shown the how to use archer message then call it now
		if not archerUnitUnlocked then
			ShowDialog("talkingheads/full/guard","Command the archers to attack the pikemen from range")
			archerUnitUnlocked = true
			--ShowUnitGuideArrow(archerUnit[1])
		else
			-- put in move orders for all remaining archer units
			for i,trigger in pairs(triggerList) do
				AddTargetCircle(trigger)
			end
		end
	end
end

function removeTriggerFromTriggerList(triggerToRemoveName)
	-- remove the trigger from the list
	for i,trigger in pairs(triggerList) do
		-- if this is the trigger unit then remove it from the array
		if trigger == triggerToRemoveName then
			-- remove the element from the array at point i
			table.remove(triggerList,i)
	
			-- leave the function (because we've found the right trigger)
			return
		end
	end
end